@@ -56,6 +56,8 @@
 #include <stdlib.h>
 #include <errno.h>
 #include <time.h>
+#include <sys/signalfd.h>
+#include <signal.h>
 
 #include <config-kscreenlocker.h>
 #if HAVE_SYS_PRCTL_H
@@ -66,74 +68,10 @@
 #include <sys/procctl.h>
 #endif
 
-
-/* Compatibility: accept some options from environment variables */
-#define ACCEPT_ENV
-
 #define THROTTLE 3
 
 static int havetty, sfd = -1, nullpass;
 
-static char *
-conv_legacy (ConvRequest what, const char *prompt)
-{
-    char		*p, *p2;
-    int			len;
-    char		buf[1024];
-
-    switch (what) {
-    case ConvGetBinary:
-	break;
-    case ConvGetNormal:
-	/* there is no prompt == 0 case */
-	if (!havetty)
-	    break;
-	/* i guess we should use /dev/tty ... */
-	fputs(prompt, stdout);
-	fflush(stdout);
-	if (!fgets(buf, sizeof(buf), stdin))
-	    return 0;
-	len = strlen(buf);
-	if (len && buf[len - 1] == '\n')
-	    buf[--len] = 0;
-	return strdup(buf);
-    case ConvGetHidden:
-	if (havetty) {
-#ifdef HAVE_GETPASSPHRASE
-	    p = getpassphrase(prompt ? prompt : "Password: ");
-#else
-	    p = getpass(prompt ? prompt : "Password: ");
-#endif
-	    p2 = strdup(p);
-	    memset(p, 0, strlen(p));
-	    return p2;
-	} else {
-	    if (prompt)
-		break;
-	    if ((len = read(0, buf, sizeof(buf) - 1)) < 0) {
-		message("Cannot read password\n");
-		return 0;
-	    } else {
-		if (len && buf[len - 1] == '\n')
-		    --len;
-		buf[len] = 0;
-		p2 = strdup(buf);
-		memset(buf, 0, len);
-		return p2;
-	    }
-	}
-    case ConvPutInfo:
-	message("Information: %s\n", prompt);
-	return 0;
-    case ConvPutError:
-	message("Error: %s\n", prompt);
-	return 0;
-    }
-    message("Authentication backend requested data type which cannot be handled.\n");
-    return 0;
-}
-
-
 static int
 Reader (void *buf, int count)
 {
@@ -270,6 +208,12 @@ conv_server (ConvRequest what, const char *prompt)
 	  nullpass = 1;
 	return msg;
       }
+    case ConvPutAuthSucceeded:
+    case ConvPutAuthFailed:
+    case ConvPutAuthError:
+    case ConvPutAuthAbort:
+    case ConvPutReadyForAuthentication:
+        return 0;
     case ConvPutInfo:
     case ConvPutError:
     default:
@@ -296,12 +240,10 @@ static void ATTR_NORETURN
 usage(int exitval)
 {
   message(
-	  "usage: kcheckpass {-h|[-c caller] [-m method] [-U username|-S handle]}\n"
+	  "usage: kcheckpass {-h|[-c caller] [-m method] -S handle}\n"
 	  "  options:\n"
 	  "    -h           this help message\n"
-	  "    -U username  authenticate the specified user instead of current user\n"
 	  "    -S handle    operate in binary server mode on file descriptor handle\n"
-	  "    -c caller    the calling application, effectively the PAM service basename\n"
 	  "    -m method    use the specified authentication method (default: \"classic\")\n"
 	  "  exit codes:\n"
 	  "    0 success\n"
@@ -315,21 +257,20 @@ usage(int exitval)
 int
 main(int argc, char **argv)
 {
-#ifdef HAVE_PAM
-  const char	*caller = KSCREENSAVER_PAM_SERVICE;
-#endif
   const char	*method = "classic";
   const char	*username = 0;
-#ifdef ACCEPT_ENV
   char		*p;
-#endif
   struct passwd	*pw;
-  int		c, nfd, lfd;
+  int		c, nfd;
   uid_t		uid;
-  time_t	nexttime;
   AuthReturn	ret;
-  struct flock lk;
-  char fname[64], fcont[64];
+  sigset_t signalMask;
+  int signalFd;
+  struct signalfd_siginfo fdsi;
+  ssize_t sigReadSize;
+  pid_t parentPid;
+
+  parentPid = getppid();
 
   // disable ptrace on kcheckpass
 #if HAVE_PR_SET_DUMPABLE
@@ -340,8 +281,9 @@ main(int argc, char **argv)
   procctl(P_PID, getpid(), PROC_TRACE_CTL, &mode);
 #endif
 
-#ifdef HAVE_OSF_C2_PASSWD
-  initialize_osf_security(argc, argv);
+  // prevent becoming an orphan while waiting for SIGUSR2
+#if HAVE_PR_SET_DUMPABLE
+  prctl(PR_SET_PDEATHSIG, SIGUSR2);
 #endif
 
   /* Make sure stdout/stderr are open */
@@ -360,22 +302,14 @@ main(int argc, char **argv)
 
   havetty = isatty(0);
 
-  while ((c = getopt(argc, argv, "hc:m:U:S:")) != -1) {
+  while ((c = getopt(argc, argv, "hm:S:")) != -1) {
     switch (c) {
     case 'h':
       usage(0);
       break;
-    case 'c':
-#ifdef HAVE_PAM
-      caller = optarg;
-#endif
-      break;
     case 'm':
       method = optarg;
       break;
-    case 'U':
-      username = optarg;
-      break;
     case 'S':
       sfd = atoi(optarg);
       break;
@@ -385,83 +319,95 @@ main(int argc, char **argv)
     }
   }
 
-#ifdef ACCEPT_ENV
-# ifdef HAVE_PAM
-  if ((p = getenv("KDE_PAM_ACTION")))
-    caller = p;
-# endif
-  if ((p = getenv("KCHECKPASS_USER")))
-    username = p;
-#endif  
+  if (sfd == -1) {
+      message("Only binary protocol supported\n");
+      return AuthError;
+  }
 
   uid = getuid();
-  if (!username) {
-    if (!(p = getenv("LOGNAME")) || !(pw = getpwnam(p)) || pw->pw_uid != uid)
-      if (!(p = getenv("USER")) || !(pw = getpwnam(p)) || pw->pw_uid != uid)
-        if (!(pw = getpwuid(uid))) {
-          message("Cannot determinate current user\n");
-          return AuthError;
-        }
-    if (!(username = strdup(pw->pw_name))) {
-      message("Out of memory\n");
-      return AuthError;
-    }
-  }
-
-  /*
-   * Throttle kcheckpass invocations to avoid abusing it for bruteforcing
-   * the password. This delay belongs to the *previous* invocation, where
-   * we can't enforce it reliably (without risking giving away the result
-   * before it is due). We don't differentiate between success and failure -
-   * it's not expected to have a noticeable adverse effect.
-   */
-  if ( uid != geteuid() ) {
-    sprintf(fname, "/var/run/kcheckpass.%d", uid);
-    if ((lfd = open(fname, O_RDWR | O_CREAT | O_NOFOLLOW, 0600)) < 0) {
-      message("Cannot open lockfile\n");
-      return AuthError;
-    }
-
-    lk.l_type = F_WRLCK;
-    lk.l_whence = SEEK_SET;
-    lk.l_start = lk.l_len = 0;
-    if (fcntl(lfd, F_SETLKW, &lk)) {
-      message("Cannot obtain lock\n");
-      return AuthError;
-    }
-
-    if ((c = read(lfd, fcont, sizeof(fcont)-1)) > 0 &&
-        (fcont[c] = '\0', sscanf(fcont, "%ld", &nexttime) == 1))
-    {
-      time_t ct = time(0);
-      if (nexttime > ct && nexttime < ct + THROTTLE)
-        sleep(nexttime - ct);
-    }
-
-    lseek(lfd, 0, SEEK_SET);
-    write(lfd, fcont, sprintf(fcont, "%lu\n", time(0) + THROTTLE));
-
-    close(lfd);
-  }
-
-  /* Now do the fandango */
-  ret = Authenticate(
-#ifdef HAVE_PAM
-                     caller,
-#endif
-                     method,
-                     username, 
-                     sfd < 0 ? conv_legacy : conv_server);
-
-    if (ret == AuthBad) {
-      message("Authentication failure\n");
-      if (!nullpass) {
-        openlog("kcheckpass", LOG_PID, LOG_AUTH);
-        syslog(LOG_NOTICE, "Authentication failure for %s (invoked by uid %d)", username, uid);
+  if (!(p = getenv("LOGNAME")) || !(pw = getpwnam(p)) || pw->pw_uid != uid)
+    if (!(p = getenv("USER")) || !(pw = getpwnam(p)) || pw->pw_uid != uid)
+      if (!(pw = getpwuid(uid))) {
+        message("Cannot determinate current user\n");
+        return AuthError;
       }
+  if (!(username = strdup(pw->pw_name))) {
+    message("Out of memory\n");
+    return AuthError;
+  }
+
+    // setup signals
+    sigemptyset(&signalMask);
+    sigaddset(&signalMask, SIGUSR1);
+    sigaddset(&signalMask, SIGUSR2);
+    // block them
+    if (sigprocmask(SIG_BLOCK, &signalMask, NULL) == -1) {
+        message("Block signal failed\n");
+        conv_server(ConvPutAuthError, 0);
+        return 1;
+    }
+    signalFd = signalfd(-1, &signalMask, SFD_CLOEXEC);
+    if (signalFd == -1) {
+        message("Signal fd failed\n");
+        conv_server(ConvPutAuthError, 0);
+        return 1;
+    }
+    // now lets block on the fd
+    for (;;) {
+        conv_server(ConvPutReadyForAuthentication, 0);
+        sigReadSize = read(signalFd, &fdsi, sizeof(struct signalfd_siginfo));
+        if (sigReadSize != sizeof(struct signalfd_siginfo)) {
+            message("Read wrong size\n");
+            return 1;
+        }
+        if (fdsi.ssi_signo == SIGUSR1) {
+            if (fdsi.ssi_pid != parentPid) {
+                message("signal from wrong process\n");
+                continue;
+            }
+            /* Now do the fandango */
+            ret = Authenticate(method,
+                                username,
+                                conv_server);
+
+            if (ret == AuthBad) {
+                message("Authentication failure\n");
+                if (!nullpass) {
+                    openlog("kcheckpass", LOG_PID, LOG_AUTH);
+                    syslog(LOG_NOTICE, "Authentication failure for %s (invoked by uid %d)", username, uid);
+                }
+            }
+            switch (ret) {
+                case AuthOk:
+                    conv_server(ConvPutAuthSucceeded, 0);
+                    break;
+                case AuthBad:
+                    conv_server(ConvPutAuthFailed, 0);
+                    break;
+                case AuthError:
+                    conv_server(ConvPutAuthError, 0);
+                    break;
+                case AuthAbort:
+                    conv_server(ConvPutAuthAbort, 0);
+                default:
+                    break;
+            }
+            if (uid != geteuid()) {
+                // we don't support multiple auth for setuid kcheckpass
+                break;
+            }
+        } else if (fdsi.ssi_signo == SIGUSR2) {
+            if (fdsi.ssi_pid != parentPid) {
+                message("signal from wrong process\n");
+                continue;
+            }
+            break;
+        } else {
+            message("unexpected signal\n");
+        }
     }
 
-  return ret;
+  return 0;
 }
 
 void
